const $isset = (flag: number, sequence = 0) => (sequence & flag) === flag
const $set = (flag: number, sequence = 0) => sequence | flag
const $unset = (flag: number, sequence = 0) => sequence & ~flag
const $toggle = (flag: number, sequence = 0) => sequence ^ flag

const combine = (flags: number[]) => flags.reduce((flags, flag) => flags + flag, 0)

type FlagHandler<T> = (flag: number, flags: number) => T
const createMultiFlagHandler = <T extends number | boolean>(fn: FlagHandler<T>) => 
    (flag: number | number[], sequence = 0): T => 
        fn(Array.isArray(flag) ? combine(flag) : flag, sequence) 

/**
 * Sets one or multiple flags on a sequence of bits.
 * 
 * @function
 * @param {number | number[]} flag The bits which should be set
 * @param {number} [sequence = 0] The bit sequence to set the flags on
 * @returns {number}
 * 
 * @example
 * import { set } from '@mightyplow/flags'
 *
 * const Option = {
 *     foo: 2 ** 0,
 *     bar: 2 ** 1,
 *     baz: 2 ** 2
 * }
 * 
 * const base = 0b000;
 * console.log(set([Option.foo, Option.baz], base).toSting(2)) // "101"
 */
const set = createMultiFlagHandler($set)

/**
 * Unsets one or multiple flags on a sequence of bits.
 * 
 * @function
 * @param {number | number[]} flag The bits which should be unset
 * @param {number} [sequence = 0] The bit sequence to set the flags on
 * @returns {number}
 */
const unset = createMultiFlagHandler($unset)

/**
 * Inverts one or multiple flags on a sequence of bits.
 * 
 * @function
 * @param {number | number[]} flag The bits which should be inverted
 * @param {number} [sequence = 0] The bit sequence to set the flags on
 * @returns {number}
 */
const toggle = createMultiFlagHandler($toggle)

/**
 * Check whether one or multiple flags are set. Returns only true when
 * all requested bits are set.
 * 
 * @function
 * @param {number | number[]} flag The bits which should be inverted
 * @param {number} [sequence = 0] The bit sequence to set the flags on
 * @returns {boolean}
 */
const isset = createMultiFlagHandler($isset)

export {
    isset,
    set,
    unset,
    toggle
}